from math import radians, log, tan, cos, pi
from urllib.parse import urlparse


class InvalidCoordinateForZoom(Exception):
    pass


class RasterTileManager:

    def __init__(self):
        # About Spherical Mercator
        # http://docs.openlayers.org/library/spherical_mercator.html
        self.spherical_mercator_xmax = 20037508.34
        self.spherical_mercator_ymax = 20037508.34
        self.spherical_mercator_xmin = -20037508.34
        self.spherical_mercator_ymin = -20037508.34
        self.tile_pixels_width = 256
        self.tile_pixels_height = 256

    def parse_url(self, url):
        """
        Parse out the layername, zoom, x, y and image file format from given URL.
        :param url: url of map server in format: 'http://www.someserver.com/partofurl/layername/zoom/x/y.png'
        :return: layername, x, y, z, tile image format (png/jpg)
        """
        result = urlparse(url)
        layername, z, x, ypart = result.path.rsplit("/", 4)[-4:]
        y, image_format = ypart.split(".")
        return layername, int(z), int(x), int(y), image_format

    def lonlat_to_tile(self, zoom, lon, lat):
        """
        Return the tile coordinates for a given longitude, latitude and zoom level.
        :param lon: (float) longitude in decimal degrees
        :param lat: (float) latitude in decimal degrees
        :param zoom: (int) zoom level (0-19)
        :return: tilex, tiley
        """
        assert 0 <= zoom <= 19
        lat_radians = radians(lat)
        n = 2.0 ** zoom
        tilex = int((lon + 180)/360.0 * n)
        tiley = int((1.0 - log(tan(lat_radians) + (1/cos(lat_radians)))/pi)/2.0 * n)
        return tilex, tiley

    def get_neighbor_tiles(self, zoom, tilex, tiley):
        """
        Obtain the neighboring tiles for a given tile
        :param zoom: (int) zoom level
        :param tilex: (int) tile x
        :param tiley: (int) tile y
        :return: (list) [(neighbor_tilex, neighbor_tiley), ...]
        """
        max_tiles, _ = self.tiles_per_dimension(zoom)
        neighbor_tiles = []
        for x_offset in (-1, 0, 1):
            for y_offset in (-1, 0, 1):
                if x_offset == 0 and y_offset == 0:
                    continue
                new_tilex = tilex + x_offset
                new_tiley = tiley + y_offset
                if new_tilex  < 0 or new_tiley < 0:
                    continue
                elif new_tilex >= max_tiles or new_tiley >= max_tiles:
                    continue
                neighbor_tiles.append((new_tilex, new_tiley))
        return neighbor_tiles

    def tile_sphericalmercator_extent(self, zoom, tilex, tiley):
        """
        Calculate the given tile's Spherical Mercator extent
        :param zoom: zoom level
        :param tilex: TMS tile X
        :param tiley: TMS tile Y
        :return: Tile's Spherical Mercator extent (minx, miny, maxx, maxy)
        """
        xtiles_at_zoom, ytiles_at_zoom = self.tiles_per_dimension(zoom)
        if not (0 <= tilex <= xtiles_at_zoom):
            msg = 'x({}) not less than expected xtiles_at_zoom({}) for zoom({})!'.format(tilex, xtiles_at_zoom, zoom)
            raise InvalidCoordinateForZoom(msg)
        if not (0 <= tiley <= ytiles_at_zoom):
            msg = 'y({}) not less than expected ytiles_at_zoom({}) for zoom({})!'.format(tiley, ytiles_at_zoom, zoom)
            raise InvalidCoordinateForZoom(msg)
        # note these values are expected to be the same since tiles are squares
        meters_per_xtile_dimension = (self.spherical_mercator_xmax + abs(self.spherical_mercator_xmin))/xtiles_at_zoom
        meters_per_ytile_dimension = (self.spherical_mercator_ymax + abs(self.spherical_mercator_ymin))/ytiles_at_zoom
        assert meters_per_xtile_dimension == meters_per_ytile_dimension

        lowerleft_tile_cornerx = (tilex * meters_per_xtile_dimension) - self.spherical_mercator_xmax
        lowerleft_tile_cornery = (tiley * meters_per_ytile_dimension) - self.spherical_mercator_ymax

        minx = lowerleft_tile_cornerx
        miny = lowerleft_tile_cornery
        maxx = lowerleft_tile_cornerx + meters_per_xtile_dimension
        maxy = lowerleft_tile_cornery + meters_per_ytile_dimension
        return minx, miny, maxx, maxy

    def sphericalmercator_to_pixel(self, zoom, tilex, tiley, xm, ym):
        """
        Given a specific zoom & tile location,
        Reproject spherical-mercator value to raster x/y pixel values
        :param zoom: zoom level
        :param tilex: TMS tile X
        :param tiley: TMS tile Y
        :param xm: X in Spherical Mercator (meters)
        :param ym: Y in Spherical Mercator (meters)
        :return: xp, yp (x, y raster pixel coordinates)
        """
        # get tile extents
        tile_minx, tile_miny, tile_maxx, tile_maxy = self.tile_sphericalmercator_extent(zoom, tilex, tiley)

        tile_meters_x_width = tile_maxx - tile_minx
        tile_meters_y_height = tile_maxy - tile_miny
        meters_per_xpixel = tile_meters_x_width/256
        meters_per_ypixel = tile_meters_y_height/256

        # adjust xm & ym to max/min values, if they exceed the given tile
        if xm > tile_maxx:
            xm = tile_maxx
        elif xm < tile_minx:
            xm = tile_minx
        if ym > tile_maxy:
            ym = tile_maxy
        elif ym < tile_miny:
            ym = tile_miny

        # shift sphereical-mercator origin to zero start from lower-left
        # --> shifts so lower right is 0,0
        # --> NOTE: value still in meters
        shifted_xm = xm - tile_minx
        shifted_ym = ym - tile_miny
        # invert y (for raster space)
        inverted_ym = shifted_ym - tile_meters_y_height

        # convert from meters (0 - shifted_max) to pixels (0 - 256)
        xp = shifted_xm / meters_per_xpixel
        yp = abs(inverted_ym / meters_per_ypixel)

        # make sure pixels are in the expected range.
        assert 0 <= xp <= 256
        assert 0 <= yp <= 256

        return int(xp), int(yp)

    def tiles_per_dimension(self, zoom):
        """
        Refer to http://wiki.openstreetmap.org/wiki/Slippy_map_tilenames for details
        :param zoom: Zoom level to calculate row/column numbers for
        :return: number of tile columns(x), rows(y) at zoom level
        """
        tile_count = 2**zoom
        # only support square tiles
        assert self.tile_pixels_height == self.tile_pixels_width
        return tile_count, tile_count




#!/usr/bin/env python

from distutils.core import setup

setup(
    name='tmstiler',
    packages=['tmstiler', ],
    version='0.5.1',
    license='MIT',
    description='Utilities for creating and calculating coordinates for Web Map Image tiles',
    author='Shane C0usins',
    author_email='shane.cousins@gmail.com',
    url='https://github.com/monkut/tmstiler',
    download_url='https://github.com/monkut/tmstiler/tarball/0.5.1/',
    keywords=['tiles', 'web-mapping', 'tile-map-service', 'tms'],
    classifiers=[
      'Development Status :: 5 - Production/Stable',
      'Intended Audience :: Developers',
      'License :: OSI Approved :: MIT License',
      'Programming Language :: Python :: 3',
      'Programming Language :: Python :: 3.2',
      'Programming Language :: Python :: 3.3',
      'Programming Language :: Python :: 3.4',
      'Programming Language :: Python :: 3.5',
    ],
)



#!/usr/bin/env python
"""
DjangoRasterTileLayerManager for django specific raster implementations.
Excepts that for each layer, a django model with a defined Point() field is given.
"""
import mimetypes

from django.contrib.gis.geos import Polygon, Point
from PIL import Image, ImageDraw

from .rtm import RasterTileManager


SPHERICAL_MERCATOR_SRID = 3857  # google maps projection


class RequiredConfigMissing(Exception):
    pass


class LayerNotConfigured(Exception):
    pass


class ObjectMissingExpectedMethod(Exception):
    pass


class ReferenceLegend:

    def get_color_str(self, model_instance, **kwargs):
        # TODO: Complete for proper color calculation based on model_instance value!
        # 1. calculate the color from the given model instance fields
        # 2. return a PIL supported color string.
        #    Refer to the link below for accepted color strings:
        #    http://pillow.readthedocs.org/en/latest/reference/ImageColor.html?highlight=hsl#color-names
        return "hsl(0,100%,50%)"  # pure red


class DjangoRasterTileLayerManager(RasterTileManager):
    LEGEND_REQUIRED_METHODS = ("get_color_str", )
    VALID_POINT_POSITIONS = ("upperleft",
                             "upperright",
                             "lowerleft",
                             "lowerright",
                             "center")
    VALID_WMS_TYPES = ("TMS", )
    LAYER_CONFIG_REQUIRED_KEYS = ("pixel_size",
                                  "point_position",
                                  "model_queryset",
                                  "model_point_fieldname",
                                  "model_value_fieldname",
                                  "legend_instance")
    LAYER_CONFIG_DEFAULTS = {"model_value_fieldname": "value",
                             "round_pixels": False,
                             "wms_type": "TMS"}

    def __init__(self, layers_config):
        """
        :param layers_config:
            { <layer name>: {
                "pixel_size":<pixel area size in meters>, # this is the raster pixel or bin size in meters
                "point_position":<pixel position represented by model point>,
                "model_queryset": <django model queryset object with model containing point & value fields>,
                "model_point_fieldname": <point fieldname>,
                "model_value_fieldname": <value fieldname>,
                "round_pixels": False,
                "legend_instance": <legend object instance with 'get_color_str()' method, for pixel color calculation>,
                 },
           }
        """
        # check incoming layer config values
        for layer_name, config_values in layers_config.items():
            if not all(required_config in config_values for required_config in self.LAYER_CONFIG_REQUIRED_KEYS):
                msg = "Given layer config missing required values! Expected values: {}".format(self.LAYER_CONFIG_REQUIRED_KEYS)
                raise RequiredConfigMissing(msg)
            assert config_values["point_position"] in self.VALID_POINT_POSITIONS
            if not all(callable(getattr(config_values["legend_instance"], m)) for m in self.LEGEND_REQUIRED_METHODS):
                msg = "given 'legend_instance' object does not have a defined '{}' method!".format(self.LEGEND_REQUIRED_METHODS)
                raise ObjectMissingExpectedMethod(msg)

            # set optional defaults
            for config_fieldname, config_default in self.LAYER_CONFIG_DEFAULTS.items():
                if config_fieldname not in config_values:
                    config_values[config_fieldname] = config_default
        self.layers_config = layers_config

        # initialize base-class variables
        super().__init__()

    def _adjust_point_to_upperleft(self, layername, point_object):
        """
        Adjust point so that it represents the upper-left coord for defined pixel size
        :param layername: Defined in layers_config on initial instantiation.
            needed to retrieve 'point_position' and 'pixel_size' for layer
        :param point_object: Django Point Object to be adjusted
        :return: Adjusted Point Object
        """
        layer_config = self.layers_config[layername]
        point_position = layer_config["point_position"]
        pixel_size = layer_config["pixel_size"]
        x, y = point_object
        if point_position == "upperright":
            x -= pixel_size
        elif point_position == "lowerright":
            x -= pixel_size
            y -= pixel_size
        elif point_position == "lowerleft":
            y -= pixel_size
        elif point_position == "center":
            x -= point_position/2.0
            y -= pixel_size/2.0

        return Point(x, y, srid=point_object.srid)

    def get_tile(self, layername, zoom, tilex, tiley, extension=".png"):
        """
        :param layername: Needed to retrieve layer specific configuration
        :param zoom: Zoom Level
        :param tilex: tile x value (upper left starts at 0)
        :param tiley: tile y value (upper left starts at 0)
        :param extension: image extension type
        :return: (<mimetype>, <resulting tile image object>)
        """
        layer_config = self.layers_config.get(layername, None)
        if not layer_config:
            raise LayerNotConfigured("layers_config[{}] not found in: {}".format(layername, str(self.layers_config.keys())))

        # get tile extents in SPHERICAL_MERCATOR_SRID
        # (xmin, ymin, xmax, ymax)
        tile_xmin, tile_ymin, tile_xmax, tile_ymax = self.tile_sphericalmercator_extent(zoom, tilex, tiley)
        tile_bbox = Polygon.from_bbox((tile_xmin, tile_ymin, tile_xmax, tile_ymax))
        tile_bbox.srid = SPHERICAL_MERCATOR_SRID
        # expand tile_bbox by 1 pixel(bin_size) to assure edge data is included
        buffered_bbox = tile_bbox.buffer(layer_config["pixel_size"], quadsegs=2)

        # start drawing each block
        tile_image = Image.new("RGBA",
                               (self.tile_pixels_width, self.tile_pixels_height),
                               (255, 255, 255, 0))
        draw = ImageDraw.Draw(tile_image)

        # get layer legend instance
        legend = layer_config["legend_instance"]

        # get & process pixel data in attached model
        kwargs = {"{}__within".format(layer_config["model_point_fieldname"]): buffered_bbox, }
        queryset = layer_config["model_queryset"]
        model_instance_pixel_data = queryset.filter(**kwargs)
        for model_instance in model_instance_pixel_data:
            color_str = legend.get_color_str(model_instance,
                                             model_value_fieldname=layer_config["model_value_fieldname"])
            model_point = getattr(model_instance, layer_config["model_point_fieldname"])
            # pixel x, y expected to be in spherical-mercator
            # attempt to transform, note if srid is not defined this will generate an error
            if model_point.srid != SPHERICAL_MERCATOR_SRID:
                model_point.transform(SPHERICAL_MERCATOR_SRID)

            # adjust to upper-left/nw
            upperleft_point = self._adjust_point_to_upperleft(layername, model_point)

            # (xmin, ymin, xmax, ymax)
            sphericalmercator_bbox = (upperleft_point.x ,
                                      upperleft_point.y - layer_config["pixel_size"],
                                      upperleft_point.x + layer_config["pixel_size"],
                                      upperleft_point.y)
            # transform pixel spherical-mercator coords to image pixel coords
            sphericalmercator_poly = Polygon.from_bbox(sphericalmercator_bbox)
            if layer_config["round_pixels"]:
                buffer_size = layer_config["pixel_size"] / 2.0
                sphericalmercator_poly = sphericalmercator_poly.centroid.buffer(buffer_size)

            poly_coords = []
            for sm_x, sm_y in sphericalmercator_poly.coords[0]:
                px, py = self.sphericalmercator_to_pixel(zoom, tilex, tiley, sm_x, sm_y)
                poly_coords.append((px, py))

            # draw pixel on tile
            draw.polygon(poly_coords, fill=color_str)

        return mimetypes.types_map.get(extension), tile_image
